"""Exposes the basic interface to interact with the serialization means."""


import dolon.impl.utils_impl as utils_impl


async def process_message(db, payload):
    await utils_impl.process_message(db, payload)


async def get_latest_trace(app_name):
    return await utils_impl.get_latest_trace(app_name)


async def get_trace(uuid, db=None):
    return await utils_impl.get_trace(uuid, db)




import asyncio
import socket
import json
import tracemalloc
import uuid

class TraceClientImpl:

    def __init__(self, app_name, host, port, *diagnostics):
        self._app_name = app_name
        self._host = host
        self._port = port
        self._diagnostics = list(diagnostics)
        self._socket = None

    def _send(self, data):
        assert self._socket
        assert isinstance(data, dict)
        data["uuid"] = self._uuid
        txt = json.dumps(data)
        self._socket.sendto(txt.encode('utf-8'), (self._host, self._port))

    async def run(self, frequency):
        while 1:
            await asyncio.sleep(frequency)
            msg = {
                "msg_type": "row",
                "row_data": [
                    await diagnostic() for diagnostic in self._diagnostics
                ]
            }
            print("Sending:", msg)
            self._send(msg)


    async def __aenter__(self):
        tracemalloc.start()
        self._uuid = str(uuid.uuid4())
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, 0)
        column_names = []
        for diagnostic in self._diagnostics:
            if hasattr(diagnostic, "__name__"):
                column_names.append(diagnostic.__name__)
            elif hasattr(diagnostic, "__class__"):
                column_names.append(diagnostic.__class__.__name__)
            else:
                raise ValueError

        msg = {
            "msg_type": "create_trace_run",
            "app_name": self._app_name,
            "column_names": column_names
        }
        self._send(msg)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self._socket:
            self._socket.close()
            self._socket = None
            tracemalloc.stop()



"""Wraps db connection pool within a context manager."""

import os

import asyncpg

# Applicable when env variable HOST not set; points to host machine
# when the application lives within a docker container.
_DEFAULT_HOST = "172.17.0.1"


class DbConnectionImpl:
    """Wraps db connection pool within a context manager."""

    _conn_pool = None

    _PREFETCH_SIZE = 100

    def __init__(self, min_size=1, max_size=10, conn_str=None):
        self._min_size = min_size
        self._max_size = max_size
        self._conn_str = conn_str

    def _get_conn_str(self):
        """Returns the database connection string.

        :return: The database connection string.
        :rtype: str
        """
        if self._conn_str:
            return self._conn_str
        else:
            user = os.environ["POSTGRES_USER"]
            password = os.environ["POSTGRES_PASSWORD"]
            db = os.environ["POSTGRES_DB"]
            host = os.environ.get("HOST", _DEFAULT_HOST)

            return f'postgresql://{user}:{password}@{host}:5432/{db}'

    def get_conn_pool(self):
        return self._conn_pool

    async def __aenter__(self):
        self._conn_pool = await asyncpg.create_pool(
            self._get_conn_str(),
            min_size=self._min_size,
            max_size=self._max_size,
            statement_cache_size=0,
            max_inactive_connection_lifetime=10,
            max_queries=1000
        )
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self._conn_pool:
            print("Closing connection")
            await self._conn_pool.close()
            self._conn_pool = None

    async def execute_query(self, sql, query_args=None):
        """Executes the passed in sql statement.

        :raises asyncpg.exceptions.PostgresError
        """
        async with self._conn_pool.acquire() as connection:
            statement = await connection.prepare(sql)
            args = query_args or []
            async with connection.transaction():
                async for record in statement.cursor(
                        *args, prefetch=self._PREFETCH_SIZE):
                    yield dict(record)



"""The trace client to record diagnostics."""

import asyncio
import tracemalloc

import dolon.impl.db_conn_impl as db_conn_impl
import dolon.impl.db_stats as db_stats
import dolon.impl.trace_client_impl as trace_client_impl


async def start_tracer(app_name, frequency, host, port, *diagnostics):
    """Starts a tracer.

    Any exception that will be caught will stop the application.

    :param str app_name: The application name that will tag the trace.
    :param int frequency: The sleep interval between consecutive traces.
    :param int port: The port to listen for messages.
    :param diagnostics: The diagnostic functions to record.
    """
    try:
        tc = trace_client_impl.TraceClientImpl(
            app_name,
            host,
            port,
            *diagnostics
        )
        async with tc:
            await tc.run(frequency)
    except Exception as ex:
        print(ex)
        exit(-1)


class PostgresDiagnostics(db_conn_impl.DbConnectionImpl):
    """Wraps memory diagnostics within an async context manager.

    :ivar str conn_str: The connection string to the database.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    async def live_msgs(self):
        """Returns the number of live messages in db."""
        return await db_stats.live_msgs_in_db.get_value(self)

    async def dead_msgs(self):
        """Returns the number of dead messages in db."""
        return await db_stats.dead_msgs_in_db.get_value(self)

    async def idle(self):
        """Returns the number of idle sessions in db."""
        return await db_stats.idle_in_db.get_value(self)

    async def dn_conn(self):
        """Returns the number of connections in db."""
        return await db_stats.conn_count_in_db.get_value(self)


class MemoryDiagnostics:
    """Wraps memory diagnostics within an async context manager."""

    async def __aenter__(self):
        """Enters the context."""
        tracemalloc.start()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Exits from the context."""
        tracemalloc.stop()

    async def mem_allocation(self):
        """Returns the size of the allocated memory in Mb."""
        current, peak = tracemalloc.get_traced_memory()
        return current / 10 ** 6


async def active_tasks():
    """Returns the number of active tasks."""
    return len([task for task in asyncio.Task.all_tasks() if not task.done()])
